import unittest
from hydrobox.utils.decorators import accept, enforce
import numpy as np
from matplotlib.axes import SubplotBase
import matplotlib.pyplot as plt


class TestAcceptDecorator(unittest.TestCase):

    def test_all_native(self):
        """
        Test all native python functionality.

        :return:
        """
        # define test f
        @accept(a=int, b=float, c=bool, d=tuple, e=list, g=dict)
        def f(a, b, c, d, e, g):
            pass

        # call it
        try:
            f(42, 42.3, True, (1,2), [1,2], {'foo':'bar'})
        except TypeError:
            self.fail('TypeError raised by accept decorator')

        # test failure
        with self.assertRaises(TypeError):
            f(42.3, 42.3, True, (1,2), [1,2], {'foo':'bar'})
        with self.assertRaises(TypeError):
            f(42, 42, True, (1, 2), [1, 2], {'foo': 'bar'})

    def test_fail_native(self):
        """
        Test if wrong type raises Type error.

        :return:
        """
        # define test f
        @accept(a=int)
        def f(a):
            pass

        # call it
        with self.assertRaises(TypeError):
            f(42.3)

    def test_callable(self):
        """
        Test if a callable is handled correctly.

        :return:
        """
        # define test
        func = lambda x: x
        @accept(a='f', b='lambda', c='callable')
        def f(a, b, c):
            pass

        # call it
        try:
            f(func, func, func)
        except TypeError:
            self.fail('TypeError raised by accept decorator')

        # test failure
        with self.assertRaises(TypeError):
            f(5,func,func)

    def test_none_values(self):
        """
        Test if None values are handled correctly.

        :return:
        """
        @accept(a='None', b=(int, 'None'))
        def f(a, b):
            pass

        # call it
        try:
            f(None, 5)
            f(None, None)
        except TypeError:
            self.fail('TypeError raised by accept decorator')

        # test failure
        with self.assertRaises(TypeError):
            f(5, 42.3)

    def test_matplotlib_numpy(self):
        """
        Test it a matplotlib Subplot and numpy ndarray are handled correctly.

        :return:
        """
        @accept(a=np.ndarray, b=SubplotBase, c=(np.ndarray, SubplotBase))
        def f(a,b,c):
            pass

        # create obj.
        arr = np.zeros(10)
        ax = plt.plot([1,2,3], [4,5,6])[0].axes

        # call it
        try:
            f(arr, ax, arr)
            f(arr, ax, ax)
        except TypeError:
            self.fail('TypeError raised by accept decorator')

        # test failure
        with self.assertRaises(TypeError):
            f(ax, arr, arr)
        with self.assertRaises(TypeError):
            f(arr, arr, arr)
        with self.assertRaises(TypeError):
            f(arr, ax, 5)


class TestEnforceDecorator(unittest.TestCase):

    def test_int(self):
        """
        Test the int type cast

        :return:
        """
        @enforce(a=int, b=int)
        def f(a, b=5):
            return a, b

        # assert positional
        self.assertEqual(f(4.2, 4.9), (4, 4))
        # assert default
        self.assertEqual(f(4.2), (4, 5))
        # assert with cast fail
        self.assertEqual(f(4.2, 'mystring'), (4, 'mystring'))
        # assert keyword argument
        self.assertEqual(f(4.2, b=7.5), (4, 7))
        # assert bool
        self.assertEqual(f(True, False), (1, 0))
        # assert string
        self.assertEqual(f('5', '4'), (5, 4))

    def test_float(self):
        """
        Test the float type cast

        :return:
        """
        @enforce(a=float, b=float)
        def f(a, b=5.5):
            return a, b

        # assert positional
        self.assertEqual(f(4, 3), (4.0, 3.0))
        # assert default
        self.assertEqual(f(4), (4, 5.5))
        # assert with cast fail
        self.assertEqual(f(4, 'mystring'), (4.0, 'mystring'))
        # assert keyword argument
        self.assertEqual(f(4, b=7), (4.0, 7.0))
        # assert bool
        self.assertEqual(f(True, False), (1.0, 0.0))
        # assert string
        self.assertEqual(f('5', '4.6'), (5.0, 4.6))

    def test_bool(self):
        """
        Test the bool type cast

        :return:
        """
        @enforce(a=bool, b=bool)
        def f(a, b=True):
            return a, b

        # assert positional
        self.assertEqual(f(0, 0), (False, False))
        # assert default
        self.assertEqual(f(1), (True, True))
        # assert with cast fail
        self.assertEqual(f(4, 'mystring'), (True, True))
        # assert keyword argument
        self.assertEqual(f(4, b=0.0), (True, False))



if __name__ == '__main__':
    unittest.main()



"""
Tools for extreme value statistics on discharge measurements.

"""
from hydrobox.utils.decorators import accept
from matplotlib.axes import SubplotBase
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import rankdata


@accept(x=np.ndarray,log=bool, plot=bool, non_exceeding=bool, ax=('None', SubplotBase))
def flow_duration_curve(x, log=True, plot=True, non_exceeding=True, ax=None, **kwargs):
    """
    Calculate and draw a flow duration curve from the discharge measurements.

    All oberservations will be ordered and the Weibull empirical probability will be calculated.
    The ordered probabilities are plotted as a flow duration curve.

    In case x.ndim > 1, the function will be called iterativly along axis 0.

    :param x:   numpy.ndarray of discharge measurements
    :param log:  bool, if True plot on loglog axis, ignored when plot is False
    :param plot: bool, if False not plotting, returning the result instead
    :param non_exceeding: bool, if true use non-exceeding probalilities
    :param ax: matplotlib Subplot object, if not None, will plot into that instance
    :param kwargs: will be passed to the matplotlib.pyplot.plot function
    :return:
    """
    if x.ndim > 1:
        raise NotImplementedError('This is not implemented yet.')

    # calculate the ranks
    ranks = rankdata(x, method='average')

    # calculate weibull pdf
    N = x.size

    # calculate probabilities
    p = np.fromiter(map(lambda r: r / (N + 1), ranks), dtype=np.float)

    # create sorting index
    if non_exceeding:
        index = np.argsort(p)
    else:
        index = np.argsort(p)[::-1]

    if not plot:
        return p[index]

    # generate an Axes, if ax is None
    if ax is None:
        fig, ax = plt.subplots(1,1)

    # plot
    # set some defaults
    kwargs.setdefault('linestyle', '-')
    kwargs.setdefault('color', 'b')
    ax.plot(x[index], p[index], **kwargs)
    #label
    ax.set_xlabel('discharge [m3/s]')
    ax.set_ylabel('%sexceeding prob.' % ('non-' if non_exceeding else ''))
    # handle loglog
    if log:
        ax.loglog()
    else:
        ax.set_ylim((-0.05, 1.1))
        ax.set_xlim(np.nanmin(x) *0.98, np.nanmax(x) * 1.02)
    ax.set_title('%sFDC' % ('loglog ' if log else ''))
    ax.grid(which='both' if log else 'major')

    return ax




from setuptools import setup, find_packages

with open('requirements.txt') as fs:
    REQUIREMENTS = fs.read().strip().split('\n')

with open('VERSION.txt') as fs:
    VERSION = fs.read().strip()

def readme():
    with open('README.rst') as fs:
        return fs.read()

setup(name='hydrobox',
      version=VERSION,
      license='MIT',
      description='Hydrological toolbox build on top of scipy and pandas',
      long_description=readme(),
      author='Mirko Maelicke',
      author_email='mirko.maelicke@kit.edu',
      install_requires=REQUIREMENTS,
      test_suite='nose.collector',
      tests_require=['nose'],
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False
)


import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal
from hydrobox.discharge.stats import flow_duration_curve

# some result container
PROBCALC = np.array([0.04761905, 0.0952381, 0.14285714, 0.19047619, 0.23809524, 0.28571429,  0.33333333,
                     0.38095238, 0.42857143, 0.47619048, 0.52380952, 0.57142857,  0.61904762,  0.66666667,
                     0.71428571, 0.76190476, 0.80952381, 0.85714286, 0.9047619,  0.95238095])


class TestFlowDurationCurve(unittest.TestCase):

    def setUp(self):
        """
        Set up some random data.
        :return:
        """
        np.random.seed(42)
        self.gamma = np.random.gamma(2, 2, size=20)

    def test_probability_calculation(self):
        assert_array_almost_equal(flow_duration_curve(self.gamma, plot=False), PROBCALC)

    def test_probability_inverse(self):
        assert_array_almost_equal(flow_duration_curve(self.gamma, plot=False, non_exceeding=False), PROBCALC[::-1])


if __name__ == '__main__':
    unittest.main()



"""
These decorators can be used to annotate the tool functions.
Most decorators will check the argument types or input data.
"""
from functools import wraps


def accept(**types):
    """
    Decorator to used to define accepted argument types for the toolbox functions.

    Usage
    -----

    .. code-block:: python

      @accept(foo=str, bar=(int,float))
      def f(foo, bar):
        pass

    :param types: arguments to the decorated function and the allowed types
    :return: decorated function
    """
    def decorator(f):
        # dig down to the original wrapped function in case more than one decorator was used
        _f = f
        while hasattr(_f, '__wrapped__'):
            _f = getattr(_f, '__wrapped__')

        # get the code, name and argnames
        code = _f.__code__
        fname = _f.__name__
        names = code.co_varnames[:code.co_argcount]

        @wraps(f)
        def decorated(*args, **kwargs):
            for argname, argtype in types.items():
                # check for argnames in kargs
                if argname in kwargs:
                    argval = kwargs.get(argname)
                else:
                    try:
                        argval = args[names.index(argname)]
                    except IndexError:
                        # TODO: Turn this into a develop level log
                        #print('DevelWarning: arg %s not passed by function %s. (Maybe default?).' % (argname, fname))
                        continue

                # check type
                if argtype == 'f' or argtype == 'callable' or argtype == 'lambda':
                    if not callable(argval):
                        raise TypeError('%s(...): arg %s: type is %s, but shall be callable.' % (fname, argname, type(argval)))
                    else:
                        continue

                elif argval is None:
                    if not (argtype == 'None' or (isinstance(argtype, (list, tuple)) and 'None' in argtype)):
                        raise TypeError('%s(...): arg %s: is None, must be %s.' %(fname, argname, argtype))
                    else:
                        continue

                # check if there is a None in argtype
                if isinstance(argtype, (tuple, list)) and 'None' in argtype:
                    argtype = list(argtype)
                    argtype.remove('None')
                    argtype = tuple(argtype)

                if not isinstance(argval, argtype):
                    raise TypeError("%s(...): arg %s: type is %s, must be %s." %
                                    (fname, argname, type(argval), argtype))

            # all checkes passed
            return f(*args, **kwargs)
        return decorated
    return decorator


def enforce(**types):
    """
    Decorator to used to define enforcing of argument type casts for the toolbox functions.

    In case a cast on one of the arguments raises a :py:class: `ValueError`, the cast will be ignored.
    Therefore it might make sense to combine a enforce decorator with a :py:func: `hydrobox.utils.decorators.accept`
    decorator in order to accept only the enforced type. In this case the
    :py:func: `hydrobox.utils.decorators.accept` decorator raises a :py:class: `ValueError`.

    Usage
    -----

    .. code-block:: python

      @accept(foo=str)
      def f(foo):
        return 'Result: %s' % foo

    The f function can now be called and foo will be casted to a string

    .. code-block:: python

    a = f(5)
    print(type(a))

    .. code-block:: bash

    >> <class 'str'>

    :param types: arguments to the decorated function and the desired types
    :return: decorated function
    """
    def decorator(f):
        # dig down to the original wrapped function in case more than one decorator was used
        _f = f
        while hasattr(_f, '__wrapped__'):
            _f = getattr(_f, '__wrapped__')

        # get the code, name and argnames
        code = _f.__code__
        fname = _f.__name__
        names = code.co_varnames[:code.co_argcount]

        @wraps(f)
        def decorated(*args, **kwargs):
            def caster(value, destination):
                try:
                    return destination(value)
                except (ValueError, Exception):
                    # TODO: On develop log note that the cast failed
                    return value

            # turn args into a list
            args = list(args)

            # cast all defined types
            for argname, argtype in types.items():
                if argname in kwargs:
                    argval = kwargs.get(argname)
                    kwargs.update({argname: caster(argval, argtype)})
                else:
                    try:
                        index = names.index(argname)
                        args[index] = caster(args[index], argtype)
                    except IndexError:
                        # TODO: Turn this into a develop level log
                        continue

            return f(*args, **kwargs)
        return decorated
    return decorator
